-- read from a grid with possible wrapping.
function getfromgrid(grid, x, y)
	wrappedx, wrappedy = ((x - 1) % grid.size) + 1, ((y - 1) % grid.size) + 1 
	if (not grid.wrap) and ((wrappedx ~= x) or (wrappedy ~= y)) then return {color = -1, stage = 0} end -- out of bounds!
	return grid[wrappedx][wrappedy]
end

-- resize a grid.
function resizegrid(grid, newsize)
	local oldsize = grid.size
	if newsize < oldsize then
		for _ = oldsize + 1, newsize do table.remove(grid, oldsize + 1) end
		for i = 1, newsize do
			for _ = oldsize + 1, newsize do table.remove(grid[i], oldsize + 1) end
		end
	elseif newsize > grid.size then
		for x = 1, newsize do
			grid[x] = grid[x] or {}
			for y = 1, newsize do
				grid[x][y] = grid[x][y] or {color = -1, stage = 0}
			end
		end
	end
	grid.size = newsize
end
	

-- make a fresh grid. Given the optional function parameter, call that function for each cell of the grid. Otherwise, just makes a blank one.
function makegrid(size, wrap, f)
	f = f or (function(x, y) return {color = -1, stage = 0} end)
	local grid = {}
	grid.size, grid.wrap = size, wrap
	for x = 1, size do
		grid[x] = {}
		for y = 1, size do
			grid[x][y] = f(x, y)
		end
	end
	return grid
end

function love.load() -- initialization stuff.
	gamedata = {} -- game data stored here. This stuff doesn't change during execution.
	gamedata.chimes = {}
	for speed = 0, 9 do
		gamedata.chimes[speed] = {}
		for note = 1, 6 do
			gamedata.chimes[speed][note] = love.sound.newSoundData(string.format("s%dn%d.wav", speed, note))
		end
	end
	gamedata.shapesize = 240
	gamedata.shapes = {}
	gamedata.icons = {}
	for color = 0, 9 do
		gamedata.shapes[color] = {}
		for stage = 1, 4 do
			gamedata.shapes[color][stage] = love.graphics.newImage(string.format("C%dS%d.png", color, stage))
		end
		gamedata.icons[color] = love.image.newImageData(string.format("C%dS3.png", color))
	end
	gamedata.keynums = {}
	for i = 0, 9 do
		gamedata.keynums[tostring(i)] = i
		gamedata.keynums[string.format("kp%d", i)] = i
	end
	gamedata.speedscale = 0.75
	gamedata.volumescale = 0.25
	gamedata.colors = {{r = 1, g = 0.5, b = 0}, {r = 1, g = 1, b = 0}, {r = 0, g = 0.75, b = 0}, {r = 0.25, g = 0.25, b = 1}, {r = 1, g = 0, b = 1}, {r = 0.75, g = 0.75, b = 0.75}, {r = 0.25, g = 0.25, b = 0.25}, {r = 0, g = 1, b = 1}, {r = 0.5, g = 0.25, b = 0}}
	gamedata.colors[0] = {r = 1, g = 0, b = 0}

	gamestate = {} -- game state stored here. This stuff does change during execution.
	gamestate.speedsetting = 5
	gamestate.speed = gamedata.speedscale / gamestate.speedsetting
	gamestate.sincelastupdate = 0
	gamestate.volume = 5
	love.audio.setVolume(gamestate.volume / 9 * gamedata.volumescale)
	gamestate.inputmode = "speed"
	gamestate.rng = love.math.newRandomGenerator(os.time())
	gamestate.randomgrid = function(x, y)
		local stage = gamestate.rng:random(9) - 7
		return (stage > 0) and {color = gamestate.rng:random(0, 9), stage = stage} or {color = -1, stage = 0}
	end
	local startinggridsize = 16
	gamestate.grid = makegrid(startinggridsize, true, gamestate.randomgrid)
	gamestate.oldgrid = makegrid(startinggridsize, true)
	gamestate.manualupdate = true
	gamestate.alive = true
	gamestate.clickcolor = 3
	do
		local multiplierx = 80
		local multipliery = multiplierx * 3 / 4
		local x, y = love.window.getDesktopDimensions()
		local maxx, maxy = math.floor(x / multiplierx), math.floor(y / multipliery)
		local minmax = math.min(maxx, maxy) - 2
		gamestate.windowx, gamestate.windowy = multiplierx * minmax, multipliery * minmax
	end		
	gamestate.updateside = true
	
	love.window.setTitle("Tetracrystals of Veluria")
	love.window.setMode(gamestate.windowx, gamestate.windowy, {resizable = true, centered = false, minwidth = 320, minheight = 240, msaa = 4})
	love.window.setIcon(gamedata.icons[gamestate.clickcolor])
end

function love.update(dt)
	collectgarbage()
	gamestate.sincelastupdate = gamestate.sincelastupdate + dt
	if (gamestate.inputmode == "speed") and ((gamestate.sincelastupdate >= gamestate.speed) or gamestate.manualupdate) then
		-- advance a generation.
		gamestate.alive = false
		local oldgrid, newgrid = gamestate.grid, gamestate.oldgrid
		for x = 1, oldgrid.size do
			for y = 1, oldgrid.size do
				-- the rules of advancement: 
				-- 1) if a crystal already exists at less than stage 4, advance it one stage.
				-- 2) if a crystal does not exist, or if the crystal is at stage 4 and about to disappear, create a stage 1 crystal if and only if exactly one of its neighbors is currently at stage 2 and none of its neighbors are currently at stage 3.
				newgrid[x][y] = {color = oldgrid[x][y].color, stage = 0}
				if (oldgrid[x][y].stage < 4) and (oldgrid[x][y].stage > 0) then
					newgrid[x][y] = {color = oldgrid[x][y].color, stage = oldgrid[x][y].stage + 1}
					gamestate.alive = true
				else
					local n, e, s, w = getfromgrid(oldgrid, x, y - 1), getfromgrid(oldgrid, x + 1, y), getfromgrid(oldgrid, x, y + 1), getfromgrid(oldgrid, x - 1, y)
					if (n.stage < 3) and (e.stage < 3) and (s.stage < 3) and (w.stage < 3) then
						local breedercount = ((n.stage == 2) and 1 or 0) + ((e.stage == 2) and 1 or 0) + ((w.stage == 2) and 1 or 0) + ((s.stage == 2) and 1 or 0)
						if breedercount == 1 then
							local newcolor = (n.stage == 2) and n.color or ((e.stage == 2) and e.color or ((w.stage == 2) and w.color or s.color))
							newgrid[x][y] = {color = newcolor, stage = 1}
							gamestate.alive = true
						end
					end
				end
			end
		end
		gamestate.grid, gamestate.oldgrid = newgrid, oldgrid
		gamestate.updategrid = true
		gamestate.sincelastupdate = 0
		gamestate.manualupdate = false
		if gamestate.alive then love.audio.play(love.audio.newSource(gamedata.chimes[gamestate.speedsetting][gamestate.rng:random(6)], "static")) end
	end
	if gamestate.updategrid then
		gamestate.renderedgrid = love.graphics.newCanvas(gamedata.shapesize * gamestate.grid.size, gamedata.shapesize * gamestate.grid.size)
		love.graphics.setCanvas(gamestate.renderedgrid)
		love.graphics.clear()
		for x = 1, gamestate.grid.size do
			for y = 1, gamestate.grid.size do
				local color, stage = gamestate.grid[x][y].color, gamestate.grid[x][y].stage
				if (color >= 0) and (stage > 0) then love.graphics.draw(gamedata.shapes[color][stage], (x - 1) * gamedata.shapesize, (y - 1) * gamedata.shapesize) end
			end
		end
		love.graphics.setCanvas()
		gamestate.updategrid = false
	end
	if gamestate.updateside then
		gamestate.side = love.graphics.newCanvas(math.floor(love.graphics.getWidth() / 4 + 0.5), love.graphics.getHeight(), {msaa = 4})
		love.graphics.setCanvas(gamestate.side)
		love.graphics.setFont(love.graphics.newFont("lucon.ttf", 12 * gamestate.side:getWidth() / 160))
		local pen = gamedata.colors[gamestate.clickcolor]
		love.graphics.setColor(pen.r, pen.g, pen.b)
		love.graphics.printf("Tetracrystals\nof\nVeluria", 0, 0, gamestate.side:getWidth(), "center")
		love.graphics.setColor(1, 1, 1)
		love.graphics.printf(string.format("\n\n\n\nDirections:\n\n0-9: set speed (currently %d)\nk: kill all crystals\nr: make random crystals\nn: make a random crystal\nw: turn %s wrap\n>: increase grid size\n<: decrease grid size\n(current size %d)\nclick: create crystal\nc, then 0-9: set color of crystal created (currently %d)\nv, then 0-9: set volume (currently %d)\nesc: cancel changing color or volume\nq: quit", gamestate.speedsetting, gamestate.grid.wrap and "off" or "on", gamestate.grid.size, gamestate.clickcolor, gamestate.volume), 0, 0, gamestate.side:getWidth())
		love.graphics.setCanvas()
		gamestate.updateside = false
	end
end

function love.draw()
	local gridscale = math.min(love.graphics.getWidth(), love.graphics.getHeight()) / (gamestate.grid.size * gamedata.shapesize)
	love.graphics.draw(gamestate.renderedgrid, 0, 0, 0, gridscale)
	love.graphics.draw(gamestate.side, love.graphics.getHeight())
end

function love.keypressed(key, scancode, isrepeat)
	local numberpressed = gamedata.keynums[key]
	if key == "q" then
		while love.audio.getActiveSourceCount() > 0 do end
		love.event.quit()
	elseif key == "escape" then
		gamestate.inputmode = "speed"
	elseif numberpressed and (gamestate.inputmode == "speed") then
		if (gamestate.speedsetting == 0) and (numberpressed == 0) then gamestate.manualupdate = true end
		gamestate.speedsetting = numberpressed
		gamestate.speed = (numberpressed == 0) and math.huge or (gamedata.speedscale / gamestate.speedsetting)
		gamestate.updateside = true
	elseif (key == "k") and (gamestate.inputmode == "speed") then
		gamestate.grid = makegrid(gamestate.grid.size, gamestate.grid.wrap)
		gamestate.updateside = true
		gamestate.updategrid = true
	elseif (key == "r") and (gamestate.inputmode == "speed") then
		gamestate.grid = makegrid(gamestate.grid.size, gamestate.grid.wrap, gamestate.randomgrid)
		gamestate.updateside = true
		gamestate.updategrid = true
	elseif (key == "n") and (gamestate.inputmode == "speed") then
		gamestate.grid[gamestate.rng:random(gamestate.grid.size)][gamestate.rng:random(gamestate.grid.size)] = {color = gamestate.rng:random(0, 9), stage = 1}
		gamestate.updateside = true
		gamestate.updategrid = true
	elseif (key == "w") and (gamestate.inputmode == "speed") then
		gamestate.grid.wrap = not gamestate.grid.wrap
		gamestate.updateside = true
	elseif (key == ",") and (gamestate.inputmode == "speed") and (gamestate.grid.size > 4) then
		resizegrid(gamestate.grid, gamestate.grid.size - 1)
		resizegrid(gamestate.oldgrid, gamestate.oldgrid.size - 1)
		gamestate.updateside = true
		gamestate.updategrid = true
	elseif (key == ".") and (gamestate.inputmode == "speed") and (gamestate.grid.size < 32) then
		resizegrid(gamestate.grid, gamestate.grid.size + 1)
		resizegrid(gamestate.oldgrid, gamestate.oldgrid.size + 1)
		gamestate.updateside = true
		gamestate.updategrid = true
	elseif (key == "c") and (gamestate.inputmode == "speed") then
		gamestate.inputmode = "color"
	elseif numberpressed and (gamestate.inputmode == "color") then
		gamestate.inputmode = "speed"
		gamestate.clickcolor = numberpressed
		love.window.setIcon(gamedata.icons[gamestate.clickcolor])
		gamestate.updateside = true
	elseif (key == "v") and (gamestate.inputmode == "speed") then
		gamestate.inputmode = "volume"
	elseif numberpressed and (gamestate.inputmode == "volume") then
		gamestate.inputmode = "speed"
		gamestate.volume = numberpressed
		love.audio.setVolume(gamestate.volume / 9 * gamedata.volumescale)
		gamestate.updateside = true
	else
		-- add other input modes here.
	end
end

function love.mousepressed(x, y, button)
	-- first, figure out where on the grid was clicked.
	local _ , height = love.graphics.getDimensions()
	if x < height then -- if it's not, we clicked in the directions somewhere
		local gridx, gridy = math.floor((x / height) * gamestate.grid.size) + 1, math.floor((y / height) * gamestate.grid.size) + 1
		-- place a crystal.
		if button <= 3 then
			gamestate.grid[gridx][gridy] = {color = gamestate.clickcolor, stage = button}
			gamestate.updateside = true
		end
	end		
end

function love.resize(w, h)
	-- how do we maintain our 4:3 aspect ratio?
	local fixedwidth = math.floor(h * 4.0 / 3 + 0.5)
	local fixedheight = math.floor(w * 3.0 / 4 + 0.5)
	-- did we resize horizontally, vertically, or diagonally?
	if gamestate.windowx ~= w then
		if gamestate.windowy ~= h then
			-- diagonally. Pick the change that gives the smallest adjustment.
			local widthchange = (fixedwidth * h) / (w * h)
			local heightchange = (w * fixedheight) / (w * h)
			widthchange = (widthchange < 1) and (1 / widthchange) or widthchange
			heightchange = (heightchange < 1) and (1 / heightchange) or heightchange
			gamestate.windowx = (widthchange < heightchange) and fixedwidth or w
			gamestate.windowy = (widthchange > heightchange) and fixedheight or h
		else
			-- horizontally.
			gamestate.windowx, gamestate.windowy = w, fixedheight
		end
	else
		-- vertically.
		gamestate.windowx, gamestate.windowy = fixedwidth, h
	end
	local posx, posy = love.window.getPosition()
	love.window.updateMode(gamestate.windowx, gamestate.windowy, {x = posx, y = posy})
	gamestate.updateside = true
end
